from typing import Any, Dict

from samtranslator.model import GeneratedProperty, Resource
from samtranslator.model.intrinsics import fnGetAtt, ref


class IAMRole(Resource):
    resource_type = "AWS::IAM::Role"
    property_types = {
        "AssumeRolePolicyDocument": GeneratedProperty(),
        "ManagedPolicyArns": GeneratedProperty(),
        "Path": GeneratedProperty(),
        "Policies": GeneratedProperty(),
        "PermissionsBoundary": GeneratedProperty(),
        "Tags": GeneratedProperty(),
    }

    runtime_attrs = {"name": lambda self: ref(self.logical_id), "arn": lambda self: fnGetAtt(self.logical_id, "Arn")}


class IAMManagedPolicy(Resource):
    resource_type = "AWS::IAM::ManagedPolicy"
    property_types = {
        "Description": GeneratedProperty(),
        "Groups": GeneratedProperty(),
        "PolicyDocument": GeneratedProperty(),
        "ManagedPolicyName": GeneratedProperty(),
        "Path": GeneratedProperty(),
        "Roles": GeneratedProperty(),
        "Users": GeneratedProperty(),
    }


class IAMRolePolicies:
    @classmethod
    def construct_assume_role_policy_for_service_principal(cls, service_principal: str) -> Dict[str, Any]:
        return {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Action": ["sts:AssumeRole"],
                    "Effect": "Allow",
                    "Principal": {"Service": [service_principal]},
                }
            ],
        }

    @classmethod
    def step_functions_start_execution_role_policy(cls, state_machine_arn, logical_id):  # type: ignore[no-untyped-def]
        return {
            "PolicyName": logical_id + "StartExecutionPolicy",
            "PolicyDocument": {
                "Statement": [{"Action": "states:StartExecution", "Effect": "Allow", "Resource": state_machine_arn}]
            },
        }

    @classmethod
    def stepfunctions_assume_role_policy(cls) -> Dict[str, Any]:
        return {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Action": ["sts:AssumeRole"],
                    "Effect": "Allow",
                    "Principal": {"Service": ["states.amazonaws.com"]},
                }
            ],
        }

    @classmethod
    def cloud_watch_log_assume_role_policy(cls) -> Dict[str, Any]:
        return {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Action": ["sts:AssumeRole"],
                    "Effect": "Allow",
                    "Principal": {"Service": ["apigateway.amazonaws.com"]},
                }
            ],
        }

    @classmethod
    def scheduler_assume_role_policy(cls) -> Dict[str, Any]:
        return {
            "Version": "2012-10-17",
            "Statement": [
                {"Action": ["sts:AssumeRole"], "Effect": "Allow", "Principal": {"Service": ["scheduler.amazonaws.com"]}}
            ],
        }

    @classmethod
    def lambda_assume_role_policy(cls) -> Dict[str, Any]:
        return {
            "Version": "2012-10-17",
            "Statement": [
                {"Action": ["sts:AssumeRole"], "Effect": "Allow", "Principal": {"Service": ["lambda.amazonaws.com"]}}
            ],
        }

    @classmethod
    def dead_letter_queue_policy(cls, action: Any, resource: Any) -> Dict[str, Any]:
        """Return the DeadLetterQueue Policy to be added to the LambdaRole
        :returns: Policy for the DeadLetterQueue
        :rtype: Dict
        """
        return {
            "PolicyName": "DeadLetterQueuePolicy",
            "PolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [{"Action": action, "Resource": resource, "Effect": "Allow"}],
            },
        }

    @classmethod
    def sqs_send_message_role_policy(cls, queue_arn: Any, logical_id: str) -> Dict[str, Any]:
        return {
            "PolicyName": logical_id + "SQSPolicy",
            "PolicyDocument": {"Statement": [{"Action": "sqs:SendMessage", "Effect": "Allow", "Resource": queue_arn}]},
        }

    @classmethod
    def sns_publish_role_policy(cls, topic_arn: Any, logical_id: str) -> Dict[str, Any]:
        return {
            "PolicyName": logical_id + "SNSPolicy",
            "PolicyDocument": {"Statement": [{"Action": "sns:publish", "Effect": "Allow", "Resource": topic_arn}]},
        }

    @classmethod
    def s3_send_event_payload_role_policy(cls, s3_arn: Any, logical_id: str) -> Dict[str, Any]:
        s3_arn_with_wild_card = {"Fn::Join": ["/", [s3_arn, "*"]]}
        return {
            "PolicyName": logical_id + "S3Policy",
            "PolicyDocument": {
                "Statement": [
                    {"Action": "s3:PutObject", "Effect": "Allow", "Resource": s3_arn_with_wild_card},
                    {"Action": "s3:ListBucket", "Effect": "Allow", "Resource": s3_arn},
                ]
            },
        }

    @classmethod
    def event_bus_put_events_role_policy(cls, event_bus_arn: Any, logical_id: str) -> Dict[str, Any]:
        return {
            "PolicyName": logical_id + "EventBridgePolicy",
            "PolicyDocument": {
                "Statement": [{"Action": "events:PutEvents", "Effect": "Allow", "Resource": event_bus_arn}]
            },
        }

    @classmethod
    def lambda_invoke_function_role_policy(cls, function_arn: Any, logical_id: str) -> Dict[str, Any]:
        return {
            "PolicyName": logical_id + "LambdaPolicy",
            "PolicyDocument": {
                "Statement": [{"Action": "lambda:InvokeFunction", "Effect": "Allow", "Resource": function_arn}]
            },
        }



class InvalidPluginException(Exception):
    """Exception raised when the provided plugin configuration is not valid.

    Attributes:
        plugin_name -- name of the plugin that caused this error
        message -- explanation of the error
    """

    def __init__(self, plugin_name: str, message: str) -> None:
        self._plugin_name = plugin_name
        self._message = message

    @property
    def message(self) -> str:
        return f"The {self._plugin_name} plugin is invalid. {self._message}"



from typing import Any, Dict


class SupportedResourceReferences:
    """
    Class that contains information about the resource references supported in this SAM template, along with the
    value they should resolve to. As the translator processes the SAM template, it keeps building up this
    collection which is finally used to resolve all the references in output CFN template.
    """

    def __init__(self) -> None:
        # This is a two level map like:
        # { "LogicalId": {"Property": "Value"} }
        self._refs: Dict[str, Dict[str, Any]] = {}

    def add(self, logical_id, property_name, value):  # type: ignore[no-untyped-def]
        """
        Add the information that resource with given `logical_id` supports the given `property`, and that a reference
        to `logical_id.property` resolves to given `value.

        Example:

            "MyApi.Deployment" -> "MyApiDeployment1234567890"

        :param logical_id: Logical ID of the resource  (Ex: MyLambdaFunction)
        :param property_name: Property on the resource that can be referenced (Ex: Alias)
        :param value: Value that this reference resolves to.
        :return: nothing
        """

        if not logical_id or not property_name:
            raise ValueError("LogicalId and property must be a non-empty string")

        if not value or not isinstance(value, str):
            raise ValueError("Property value must be a non-empty string")

        if logical_id not in self._refs:
            self._refs[logical_id] = {}

        if property_name in self._refs[logical_id]:
            raise ValueError(f"Cannot add second reference value to {logical_id}.{property_name} property")

        self._refs[logical_id][property_name] = value

    def get(self, logical_id, property_name):  # type: ignore[no-untyped-def]
        """
        Returns the value of the reference for given logical_id at given property. Ex: MyFunction.Alias

        :param logical_id: Logical Id of the resource
        :param property_name: Property of the resource you want to resolve. None if you want to get value of all properties
        :return: Value of this property if present. None otherwise
        """

        # By defaulting to empty dictionary, we can handle the case where logical_id is not in map without if statements
        prop_values = self.get_all(logical_id)  # type: ignore[no-untyped-call]
        if prop_values:
            return prop_values.get(property_name, None)
        return None

    def get_all(self, logical_id):  # type: ignore[no-untyped-def]
        """
        Get all properties and their values supported by the resource with given logical ID

        :param logical_id: Logical ID of the resource
        :return: Map of property names to values. None, if the logicalId does not exist
        """
        return self._refs.get(logical_id, None)

    def __len__(self) -> int:
        """
        To make len(this_object) work
        :return: Number of resource references available
        """
        return len(self._refs)

    def __str__(self) -> str:
        return str(self._refs)



from abc import ABC, abstractmethod
from collections import defaultdict
from enum import Enum
from typing import Any, Dict, List, Optional, Sequence, Union


class ExpectedType(Enum):
    MAP = ("map", dict)
    LIST = ("list", list)
    STRING = ("string", str)
    INTEGER = ("integer", int)
    BOOLEAN = ("boolean", bool)


class ExceptionWithMessage(ABC, Exception):
    @property
    @abstractmethod
    def message(self) -> str:
        """Return the exception message."""

    @property
    def metadata(self) -> Optional[Dict[str, Any]]:
        """Return the exception metadata."""


class InvalidDocumentException(ExceptionWithMessage):
    """Exception raised when the given document is invalid and cannot be transformed.

    Attributes:
        message -- explanation of the error
        metadata -- a dictionary of metadata (key, value pair)
        causes -- list of errors which caused this document to be invalid
    """

    def __init__(self, causes: Sequence[ExceptionWithMessage]) -> None:
        self._causes = list(causes)
        # Sometimes, the same error could be raised from different plugins,
        # so here we do a deduplicate based on the message:
        self._causes = list({cause.message: cause for cause in self._causes}.values())

    @property
    def message(self) -> str:
        return f"Invalid Serverless Application Specification document. Number of errors found: {len(self.causes)}."

    @property
    def metadata(self) -> Dict[str, List[Any]]:
        # Merge metadata in each exception to one single metadata dictionary
        metadata_dict = defaultdict(list)
        for cause in self.causes:
            if not cause.metadata:
                continue
            for k, v in cause.metadata.items():
                metadata_dict[k].append(v)
        return metadata_dict

    @property
    def causes(self) -> Sequence[ExceptionWithMessage]:
        return self._causes


class DuplicateLogicalIdException(ExceptionWithMessage):
    """Exception raised when a transformation adds a resource with a logical id which already exists.
    Attributes:
        message -- explanation of the error
    """

    def __init__(self, logical_id: str, duplicate_id: str, resource_type: str) -> None:
        self._logical_id = logical_id
        self._duplicate_id = duplicate_id
        self._type = resource_type

    @property
    def message(self) -> str:
        return (
            f"Transforming resource with id [{self._logical_id}] attempts to create a new"
            f' resource with id [{self._duplicate_id}] and type "{self._type}". A resource with that id already'
            " exists within this template. Please use a different id for that resource."
        )


class InvalidTemplateException(ExceptionWithMessage):
    """Exception raised when the template structure is invalid

    Attributes
        message -- explanation of the error
    """

    def __init__(self, message: str) -> None:
        self._message = message

    @property
    def message(self) -> str:
        return f"Structure of the SAM template is invalid. {self._message}"


class InvalidResourceException(ExceptionWithMessage):
    """Exception raised when a resource is invalid.

    Attributes:
        message -- explanation of the error
    """

    def __init__(
        self, logical_id: Union[str, List[str]], message: str, metadata: Optional[Dict[str, Any]] = None
    ) -> None:
        self._logical_id = logical_id
        self._message = message
        self._metadata = metadata

    def __lt__(self, other):  # type: ignore[no-untyped-def]
        return self._logical_id < other._logical_id

    @property
    def message(self) -> str:
        return f"Resource with id [{self._logical_id}] is invalid. {self._message}"

    @property
    def metadata(self) -> Optional[Dict[str, Any]]:
        return self._metadata


class InvalidResourcePropertyTypeException(InvalidResourceException):
    def __init__(
        self,
        logical_id: str,
        key_path: str,
        expected_type: Optional[ExpectedType],
        message: Optional[str] = None,
    ) -> None:
        message = message or self._default_message(key_path, expected_type)
        super().__init__(logical_id, message)

        self.key_path = key_path

    @staticmethod
    def _default_message(key_path: str, expected_type: Optional[ExpectedType]) -> str:
        if expected_type:
            type_description, _ = expected_type.value
            return f"Property '{key_path}' should be a {type_description}."
        return f"Type of property '{key_path}' is invalid."


class InvalidResourceAttributeTypeException(InvalidResourceException):
    def __init__(
        self,
        logical_id: str,
        key_path: str,
        expected_type: Optional[ExpectedType],
        message: Optional[str] = None,
    ) -> None:
        message = message or self._default_message(logical_id, key_path, expected_type)
        super().__init__(logical_id, message)

    @staticmethod
    def _default_message(logical_id: str, key_path: str, expected_type: Optional[ExpectedType]) -> str:
        if expected_type:
            type_description, _ = expected_type.value
            return f"Attribute '{key_path}' should be a {type_description}."
        return f"Type of attribute '{key_path}' is invalid."


class InvalidEventException(ExceptionWithMessage):
    """Exception raised when an event is invalid.

    Attributes:
        message -- explanation of the error
    """

    # Note: event_id should not be None, but currently there are too many
    # usage of this class with `event_id` being Optional.
    # TODO: refactor the code to make type correct.
    def __init__(self, event_id: Optional[str], message: str) -> None:
        self._event_id = event_id
        self._message = message

    @property
    def message(self) -> str:
        return f"Event with id [{self._event_id}] is invalid. {self._message}"


def prepend(exception, message, end=": "):  # type: ignore[no-untyped-def]
    """Prepends the first argument (i.e., the exception message) of the a BaseException with the provided message.
    Useful for reraising exceptions with additional information.

    :param BaseException exception: the exception to prepend
    :param str message: the message to prepend
    :param str end: the separator to add to the end of the provided message
    :returns: the exception
    """
    exception.args = exception.args or ("",)
    exception.args = (message + end + exception.args[0],) + exception.args[1:]
    return exception



try:
    from pydantic import v1 as pydantic
except ImportError:
    # Unfortunately mypy cannot handle this try/expect pattern, and "type: ignore"
    # is the simplest work-around. See: https://github.com/python/mypy/issues/1153
    import pydantic  # type: ignore

__all__ = ["pydantic"]
